@@ -4,7 +4,7 @@
 //
 //  Eureka DOOM Editor
 //
-//  Copyright (C) 2001-2009 Andrew Apted
+//  Copyright (C) 2001-2013 Andrew Apted
 //  Copyright (C) 1997-2003 Andr Majorel et al
 //
 //  This program is free software; you can redistribute it and/or
@@ -28,12 +28,18 @@
 
 #include <algorithm>
 #include <list>
+#include <string>
 
 #include "e_basis.h"
+#include "levels.h"
 #include "m_strings.h"
 
+#include "m_game.h"  // g_default_xxx
 
-extern int MadeChanges;
+// need these for the XXX_Notify() prototypes
+#include "e_cutpaste.h"
+#include "objects.h"
+#include "selectn.h"
 
 
 std::vector<Thing *>   Things;
@@ -41,17 +47,79 @@ std::vector<Vertex *>  Vertices;
 std::vector<Sector *>  Sectors;
 std::vector<SideDef *> SideDefs;
 std::vector<LineDef *> LineDefs;
-std::vector<RadTrig *> RadTrigs;
+
+
+int default_floor_h		=   0;
+int default_ceil_h		= 128;
+int default_light_level	= 176;
+int default_thing		= 2001;
+
+const char * default_floor_tex	= "FLAT1";
+const char * default_ceil_tex	= "FLAT1";
+const char * default_lower_tex	= "STARTAN3";
+const char * default_mid_tex	= "STARTAN3";
+const char * default_upper_tex	= "STARTAN3";
+
+
+static bool did_make_changes;
 
 
 string_table_c basis_strtab;
 
 
+int NumObjects(obj_type_e type)
+{
+	switch (type)
+	{
+		case OBJ_THINGS:   return NumThings;
+		case OBJ_LINEDEFS: return NumLineDefs;
+		case OBJ_SIDEDEFS: return NumSideDefs;
+		case OBJ_VERTICES: return NumVertices;
+		case OBJ_SECTORS:  return NumSectors;
+
+		default:
+			BugError("NumObjects: bad type: %d\n", (int)type);
+			return 0; /* NOT REACHED */
+	}
+}
+
+
+static void DoClearChangeStatus()
+{
+	did_make_changes = false;
+
+	Clipboard_NotifyBegin();
+	Selection_NotifyBegin();
+	 MapStuff_NotifyBegin();
+	ObjectBox_NotifyBegin();
+}
+
+static void DoProcessChangeStatus()
+{
+	if (did_make_changes)
+		MarkChanges();
+
+	Clipboard_NotifyEnd();
+	Selection_NotifyEnd();
+	 MapStuff_NotifyEnd();
+	ObjectBox_NotifyEnd();
+}
+
+
 int BA_InternaliseString(const char *str)
 {
 	return basis_strtab.add(str);
 }
 
+int BA_InternaliseShortStr(const char *str, int max_len)
+{
+	std::string goodie(str, max_len);
+
+	int result = BA_InternaliseString(goodie.c_str());
+
+	return result;
+}
+
 const char *BA_GetString(int offset)
 {
 	return basis_strtab.get(offset);
@@ -68,6 +136,18 @@ const char * Sector::CeilTex() const
 	return basis_strtab.get(ceil_tex);
 }
 
+void Sector::SetDefaults()
+{
+	floorh = default_floor_h;
+	 ceilh = default_ceil_h;
+
+	floor_tex = BA_InternaliseString(default_floor_tex);
+	 ceil_tex = BA_InternaliseString(default_ceil_tex);
+
+	light = default_light_level;
+}
+
+
 const char * SideDef::UpperTex() const
 {
 	return basis_strtab.get(upper_tex);
@@ -83,6 +163,17 @@ const char * SideDef::LowerTex() const
 	return basis_strtab.get(lower_tex);
 }
 
+void SideDef::SetDefaults(bool two_sided)
+{
+	lower_tex = BA_InternaliseString(default_lower_tex);
+	upper_tex = BA_InternaliseString(default_upper_tex);
+
+	if (two_sided)
+		mid_tex = BA_InternaliseString("-");
+	else
+		mid_tex = BA_InternaliseString(default_mid_tex);
+}
+
 
 Sector * SideDef::SecRef() const
 {
@@ -110,6 +201,58 @@ SideDef * LineDef::Left() const
 }
 
 
+bool LineDef::TouchesSector(int sec_num) const
+{
+	if (right >= 0 && SideDefs[right]->sector == sec_num)
+		return true;
+
+	if (left >= 0 && SideDefs[left]->sector == sec_num)
+		return true;
+
+	return false;
+}
+
+
+int LineDef::WhatSector(int side) const
+{
+	switch (side)
+	{
+		case SIDE_LEFT:
+			return Left() ? Left()->sector : -1;
+
+		case SIDE_RIGHT:
+			return Right() ? Right()->sector : -1;
+
+		default:
+			BugError("bad side : %d\n", side);
+			return -1;
+	}
+}
+
+
+int LineDef::WhatSideDef(int side) const
+{
+	switch (side)
+	{
+		case SIDE_LEFT:  return left;
+		case SIDE_RIGHT: return right;
+
+		default:
+			BugError("bad side : %d\n", side);
+			return -1;
+	}
+}
+
+
+double LineDef::CalcLength() const
+{
+	int dx = Start()->x - End()->x;
+	int dy = Start()->y - End()->y;
+
+	return sqrt(dx * dx + dy * dy);
+}
+
+
 //------------------------------------------------------------------------
 
 
@@ -119,7 +262,7 @@ static void RawInsertThing(int objnum, int *ptr)
 
 	Things.push_back(NULL);
 
-	for (int n = NumThings-1; n > objnum; n--)
+	for (int n = NumThings-1 ; n > objnum ; n--)
 		Things[n] = Things[n - 1];
 
 	Things[objnum] = (Thing*) ptr;
@@ -127,11 +270,11 @@ static void RawInsertThing(int objnum, int *ptr)
 
 static void RawInsertLineDef(int objnum, int *ptr)
 {
-	SYS_ASSERT(0 <= objnum && objnum <= NumThings);
+	SYS_ASSERT(0 <= objnum && objnum <= NumLineDefs);
 
 	LineDefs.push_back(NULL);
 
-	for (int n = NumLineDefs-1; n > objnum; n--)
+	for (int n = NumLineDefs-1 ; n > objnum ; n--)
 		LineDefs[n] = LineDefs[n - 1];
 
 	LineDefs[objnum] = (LineDef*) ptr;
@@ -143,7 +286,7 @@ static void RawInsertVertex(int objnum, int *ptr)
 
 	Vertices.push_back(NULL);
 
-	for (int n = NumVertices-1; n > objnum; n--)
+	for (int n = NumVertices-1 ; n > objnum ; n--)
 		Vertices[n] = Vertices[n - 1];
 
 	Vertices[objnum] = (Vertex*) ptr;
@@ -152,7 +295,7 @@ static void RawInsertVertex(int objnum, int *ptr)
 
 	if (objnum+1 < NumVertices)
 	{
-		for (int n = NumLineDefs-1; n >= 0; n--)
+		for (int n = NumLineDefs-1 ; n >= 0 ; n--)
 		{
 			LineDef *L = LineDefs[n];
 
@@ -171,7 +314,7 @@ static void RawInsertSideDef(int objnum, int *ptr)
 
 	SideDefs.push_back(NULL);
 
-	for (int n = NumSideDefs-1; n > objnum; n--)
+	for (int n = NumSideDefs-1 ; n > objnum ; n--)
 		SideDefs[n] = SideDefs[n - 1];
 
 	SideDefs[objnum] = (SideDef*) ptr;
@@ -180,7 +323,7 @@ static void RawInsertSideDef(int objnum, int *ptr)
 
 	if (objnum+1 < NumSideDefs)
 	{
-		for (int n = NumLineDefs-1; n >= 0; n--)
+		for (int n = NumLineDefs-1 ; n >= 0 ; n--)
 		{
 			LineDef *L = LineDefs[n];
 
@@ -199,7 +342,7 @@ static void RawInsertSector(int objnum, int *ptr)
 
 	Sectors.push_back(NULL);
 
-	for (int n = NumSectors-1; n > objnum; n--)
+	for (int n = NumSectors-1 ; n > objnum ; n--)
 		Sectors[n] = Sectors[n - 1];
 
 	Sectors[objnum] = (Sector*) ptr;
@@ -208,7 +351,7 @@ static void RawInsertSector(int objnum, int *ptr)
 
 	if (objnum+1 < NumSectors)
 	{
-		for (int n = NumSideDefs-1; n >= 0; n--)
+		for (int n = NumSideDefs-1 ; n >= 0 ; n--)
 		{
 			SideDef *S = SideDefs[n];
 
@@ -218,26 +361,13 @@ static void RawInsertSector(int objnum, int *ptr)
 	}
 }
 
-static void RawInsertRadTrig(int objnum, int *ptr)
-{
-	SYS_ASSERT(0 <= objnum && objnum <= NumRadTrigs);
-
-	RadTrigs.push_back(NULL);
-
-	for (int n = NumRadTrigs-1; n > objnum; n--)
-		RadTrigs[n] = RadTrigs[n - 1];
-
-	RadTrigs[objnum] = (RadTrig*) ptr;
-}
-
-
 static int * RawDeleteThing(int objnum)
 {
 	SYS_ASSERT(0 <= objnum && objnum < NumThings);
 
 	int * result = (int*) Things[objnum];
 
-	for (int n = objnum; n < NumThings-1; n++)
+	for (int n = objnum ; n < NumThings-1 ; n++)
 		Things[n] = Things[n + 1];
 
 	Things.pop_back();
@@ -245,18 +375,21 @@ static int * RawDeleteThing(int objnum)
 	return result;
 }
 
-static void RawInsert(obj_type_t objtype, int objnum, int *ptr)
+static void RawInsert(obj_type_e objtype, int objnum, int *ptr)
 {
+	did_make_changes = true;
+
+	Clipboard_NotifyInsert(objtype, objnum);
+	Selection_NotifyInsert(objtype, objnum);
+	 MapStuff_NotifyInsert(objtype, objnum);
+	ObjectBox_NotifyInsert(objtype, objnum);
+
 	switch (objtype)
 	{
 		case OBJ_THINGS:
 			RawInsertThing(objnum, ptr);
 			break;
 
-		case OBJ_LINEDEFS:
-			RawInsertLineDef(objnum, ptr);
-			break;
-
 		case OBJ_VERTICES:
 			RawInsertVertex(objnum, ptr);
 			break;
@@ -264,17 +397,17 @@ static void RawInsert(obj_type_t objtype, int objnum, int *ptr)
 		case OBJ_SIDEDEFS:
 			RawInsertSideDef(objnum, ptr);
 			break;
-		
+
 		case OBJ_SECTORS:
 			RawInsertSector(objnum, ptr);
 			break;
 
-		case OBJ_RADTRIGS:
-			RawInsertRadTrig(objnum, ptr);
+		case OBJ_LINEDEFS:
+			RawInsertLineDef(objnum, ptr);
 			break;
 
 		default:
-			nf_bug ("RawInsert: bad objtype %d", (int) objtype);
+			BugError("RawInsert: bad objtype %d", (int) objtype);
 	}
 }
 
@@ -285,7 +418,7 @@ static int * RawDeleteLineDef(int objnum)
 
 	int * result = (int*) LineDefs[objnum];
 
-	for (int n = objnum; n < NumLineDefs-1; n++)
+	for (int n = objnum ; n < NumLineDefs-1 ; n++)
 		LineDefs[n] = LineDefs[n + 1];
 	
 	LineDefs.pop_back();
@@ -299,7 +432,7 @@ static int * RawDeleteVertex(int objnum)
 
 	int * result = (int*) Vertices[objnum];
 
-	for (int n = objnum; n < NumVertices-1; n++)
+	for (int n = objnum ; n < NumVertices-1 ; n++)
 		Vertices[n] = Vertices[n + 1];
 
 	Vertices.pop_back();
@@ -308,7 +441,7 @@ static int * RawDeleteVertex(int objnum)
 
 	if (objnum < NumVertices)
 	{
-		for (int n = NumLineDefs-1; n >= 0; n--)
+		for (int n = NumLineDefs-1 ; n >= 0 ; n--)
 		{
 			LineDef *L = LineDefs[n];
 
@@ -329,7 +462,7 @@ static int * RawDeleteSideDef(int objnum)
 
 	int * result = (int*) SideDefs[objnum];
 
-	for (int n = objnum; n < NumSideDefs-1; n++)
+	for (int n = objnum ; n < NumSideDefs-1 ; n++)
 		SideDefs[n] = SideDefs[n + 1];
 
 	SideDefs.pop_back();
@@ -338,7 +471,7 @@ static int * RawDeleteSideDef(int objnum)
 
 	if (objnum < NumSideDefs)
 	{
-		for (int n = NumLineDefs-1; n >= 0; n--)
+		for (int n = NumLineDefs-1 ; n >= 0 ; n--)
 		{
 			LineDef *L = LineDefs[n];
 
@@ -359,7 +492,7 @@ static int * RawDeleteSector(int objnum)
 
 	int * result = (int*) Sectors[objnum];
 
-	for (int n = objnum; n < NumSectors-1; n++)
+	for (int n = objnum ; n < NumSectors-1 ; n++)
 		Sectors[n] = Sectors[n + 1];
 
 	Sectors.pop_back();
@@ -368,7 +501,7 @@ static int * RawDeleteSector(int objnum)
 
 	if (objnum < NumSectors)
 	{
-		for (int n = NumSideDefs-1; n >= 0; n--)
+		for (int n = NumSideDefs-1 ; n >= 0 ; n--)
 		{
 			SideDef *S = SideDefs[n];
 
@@ -380,22 +513,15 @@ static int * RawDeleteSector(int objnum)
 	return result;
 }
 
-static int * RawDeleteRadTrig(int objnum)
+static int * RawDelete(obj_type_e objtype, int objnum)
 {
-	SYS_ASSERT(0 <= objnum && objnum < NumRadTrigs);
+	did_make_changes = true;
 
-	int * result = (int*) RadTrigs[objnum];
+	Clipboard_NotifyDelete(objtype, objnum);
+	Selection_NotifyDelete(objtype, objnum);
+	 MapStuff_NotifyDelete(objtype, objnum);
+	ObjectBox_NotifyDelete(objtype, objnum);
 
-	for (int n = objnum; n < NumRadTrigs-1; n++)
-		RadTrigs[n] = RadTrigs[n + 1];
-
-	RadTrigs.pop_back();
-
-	return result;
-}
-
-static int * RawDelete(obj_type_t objtype, int objnum)
-{
 	switch (objtype)
 	{
 		case OBJ_THINGS:
@@ -413,44 +539,16 @@ static int * RawDelete(obj_type_t objtype, int objnum)
 		case OBJ_LINEDEFS:
 			return RawDeleteLineDef(objnum);
 
-		case OBJ_RADTRIGS:
-			return RawDeleteRadTrig(objnum);
-
 		default:
-			nf_bug ("RawDelete: bad objtype %d", (int) objtype);
+			BugError("RawDelete: bad objtype %d", (int) objtype);
+			return NULL; /* NOT REACHED */
 	}
 }
 
 
-static int * RawGetBase(obj_type_t objtype, int objnum)
-{
-	switch (objtype)
-	{
-		case OBJ_THINGS:
-			return (int*) Things[objnum];
-
-		case OBJ_VERTICES:
-			return (int*) Vertices[objnum];
-
-		case OBJ_SECTORS:
-			return (int*) Sectors[objnum];
-
-		case OBJ_SIDEDEFS:
-			return (int*) SideDefs[objnum];
-		
-		case OBJ_LINEDEFS:
-			return (int*) LineDefs[objnum];
-
-		case OBJ_RADTRIGS:
-			return (int*) RadTrigs[objnum];
-
-		default:
-			nf_bug ("RawGetBase: bad objtype %d", (int) objtype);
-	}
-}
-
-static void DeleteFinally(obj_type_t objtype, int *ptr)
+static void DeleteFinally(obj_type_e objtype, int *ptr)
 {
+// fprintf(stderr, "DeleteFinally: %p\n", ptr);
 	switch (objtype)
 	{
 		case OBJ_THINGS:   delete (Thing *)   ptr; break;
@@ -458,14 +556,55 @@ static void DeleteFinally(obj_type_t objtype, int *ptr)
 		case OBJ_SECTORS:  delete (Sector *)  ptr; break;
 		case OBJ_SIDEDEFS: delete (SideDef *) ptr; break;
 		case OBJ_LINEDEFS: delete (LineDef *) ptr; break;
-		case OBJ_RADTRIGS: delete (RadTrig *) ptr; break;
 
 		default:
-			nf_bug ("DeleteFinally: bad objtype %d", (int) objtype);
+			BugError("DeleteFinally: bad objtype %d", (int) objtype);
 	}
 }
 
 
+static void RawChange(obj_type_e objtype, int objnum, int field, int *value)
+{
+	int * pos = NULL;
+
+	switch (objtype)
+	{
+		case OBJ_THINGS:
+			pos = (int*) Things[objnum];
+			break;
+
+		case OBJ_VERTICES:
+			pos = (int*) Vertices[objnum];
+			break;
+
+		case OBJ_SECTORS:
+			pos = (int*) Sectors[objnum];
+			break;
+
+		case OBJ_SIDEDEFS:
+			pos = (int*) SideDefs[objnum];
+			break;
+		
+		case OBJ_LINEDEFS:
+			pos = (int*) LineDefs[objnum];
+			break;
+
+		default:
+			BugError("RawGetBase: bad objtype %d", (int) objtype);
+			return; /* NOT REACHED */
+	}
+
+	std::swap(pos[field], *value);
+
+	did_make_changes = true;
+
+	Clipboard_NotifyChange(objtype, objnum, field);
+	Selection_NotifyChange(objtype, objnum, field);
+	 MapStuff_NotifyChange(objtype, objnum, field);
+	ObjectBox_NotifyChange(objtype, objnum, field);
+}
+
+
 //------------------------------------------------------------------------
 //  BASIS API IMPLEMENTATION
 //------------------------------------------------------------------------
@@ -483,6 +622,7 @@ class edit_op_c
 {
 public:
 	char action;
+
 	byte objtype;
 	byte field;
 	byte _pad;
@@ -497,42 +637,43 @@ public:
 	{ }
 
 	~edit_op_c()
-	{
-		if (action == OP_INSERT)
-		{
-			SYS_ASSERT(ptr);
-			DeleteFinally((obj_type_t) objtype, ptr);
-		}
-		else
-		{
-			SYS_ASSERT(! ptr);
-		}
-	}
+	{ }
 
 	void Apply()
 	{
-		int * pos;
-
 		switch (action)
 		{
 			case OP_CHANGE:
-				pos = RawGetBase((obj_type_t)objtype, objnum);
-				std::swap(pos[field], value);
+				RawChange((obj_type_e)objtype, objnum, (int)field, &value);
 				return;
 
 			case OP_DELETE:
-				ptr = RawDelete((obj_type_t)objtype, objnum);
+				ptr = RawDelete((obj_type_e)objtype, objnum);
 				action = OP_INSERT;  // reverse the operation
 				return;
 
 			case OP_INSERT:
-				RawInsert((obj_type_t)objtype, objnum, ptr);
+				RawInsert((obj_type_e)objtype, objnum, ptr);
 				ptr = NULL;
 				action = OP_DELETE;  // reverse the operation
 				return;
 
 			default:
-				nf_bug("edit_op_c::Apply");
+				BugError("edit_op_c::Apply");
+		}
+	}
+
+	void Destroy()
+	{
+// fprintf(stderr, "edit_op_c::Destroy %p action = '%c'\n", this, (action == 0) ? ' ' : action);
+		if (action == OP_INSERT)
+		{
+			SYS_ASSERT(ptr);
+			DeleteFinally((obj_type_e) objtype, ptr);
+		}
+		else if (action == OP_DELETE)
+		{
+			SYS_ASSERT(! ptr);
 		}
 	}
 };
@@ -550,11 +691,22 @@ public:
 	{ }
 
 	~undo_group_c()
-	{ }
+	{
+		for (int i = (int)ops.size() - 1 ; i >= 0 ; i--)
+		{
+			ops[i].Destroy();
+		}
+	}
+
+	bool Empty() const
+	{
+		return ops.empty();
+	}
 
 	void Add_Apply(edit_op_c& op)
 	{
 		ops.push_back(op);
+
 		ops.back().Apply();
 	}
 
@@ -592,6 +744,8 @@ static void ClearUndoHistory()
 
 	for (LI = undo_history.begin(); LI != undo_history.end(); LI++)
 		delete *LI;
+
+	undo_history.clear();
 }
 
 static void ClearRedoFuture()
@@ -600,33 +754,63 @@ static void ClearRedoFuture()
 
 	for (LI = redo_future.begin(); LI != redo_future.end(); LI++)
 		delete *LI;
+
+	redo_future.clear();
 }
 
 
 void BA_Begin()
 {
 	if (cur_group)
-		nf_bug("BA_Begin called twice without BA_End\n");
+		BugError("BA_Begin called twice without BA_End\n");
 
 	ClearRedoFuture();
 
 	cur_group = new undo_group_c();
+
+	DoClearChangeStatus();
 }
 
 void BA_End()
 {
 	if (! cur_group)
-		nf_bug("BA_End called without a previous BA_Begin\n");
+		BugError("BA_End called without a previous BA_Begin\n");
 
 	cur_group->End();
 
-	undo_history.push_front(cur_group);
+	if (! cur_group->Empty())
+	{
+		undo_history.push_front(cur_group);
+	}
 
 	cur_group = NULL;
+
+	DoProcessChangeStatus();
 }
 
 
-int BA_New(obj_type_t type)
+void BA_Abort(bool keep_changes)
+{
+	if (! cur_group)
+		BugError("BA_Abort called without a previous BA_Begin\n");
+
+	cur_group->End();
+
+	if (! keep_changes && ! cur_group->Empty())
+	{
+		cur_group->ReApply();
+	}
+
+	delete cur_group;
+	cur_group = NULL;
+
+	did_make_changes  = false;
+
+	DoProcessChangeStatus();
+}
+
+
+int BA_New(obj_type_e type)
 {
 	edit_op_c op;
 
@@ -660,21 +844,18 @@ int BA_New(obj_type_t type)
 			op.ptr = (int*) new Sector;
 			break;
 
-		case OBJ_RADTRIGS:
-			op.objnum = NumRadTrigs;
-			op.ptr = (int*) new RadTrig;
-			break;
-
-		default: nf_bug("BA_New");
+		default: BugError("BA_New");
 	}
 
 	SYS_ASSERT(cur_group);
 
 	cur_group->Add_Apply(op);
+
+	return op.objnum;
 }
 
 
-void BA_Delete(obj_type_t type, int objnum)
+void BA_Delete(obj_type_e type, int objnum)
 {
 	edit_op_c op;
 
@@ -682,12 +863,12 @@ void BA_Delete(obj_type_t type, int objnum)
 	op.objtype = type;
 	op.objnum  = objnum;
 
-	// this must happen  _before_ doing the deletion (otherwise
+	// this must happen _before_ doing the deletion (otherwise
 	// when we undo, the insertion will mess up the references).
 	if (type == OBJ_SIDEDEFS)
 	{
 		// unbind sidedef from any linedefs using it
-		for (int n = NumLineDefs-1; n >= 0; n--)
+		for (int n = NumLineDefs-1 ; n >= 0 ; n--)
 		{
 			LineDef *L = LineDefs[n];
 
@@ -698,38 +879,41 @@ void BA_Delete(obj_type_t type, int objnum)
 				BA_ChangeLD(n, LineDef::F_LEFT, -1);
 		}
 	}
-
-	SYS_ASSERT(cur_group);
-
-	cur_group->Add_Apply(op);
-
-	/* these must be done after the deletion */
-
-	if (type == OBJ_VERTICES)
+	else if (type == OBJ_VERTICES)
 	{
 		// delete any linedefs bound to this vertex
-		for (int n = NumLineDefs-1; n >= 0; n--)
+		for (int n = NumLineDefs-1 ; n >= 0 ; n--)
 		{
 			LineDef *L = LineDefs[n];
 
 			if (L->start == objnum || L->end == objnum)
+			{
 				BA_Delete(OBJ_LINEDEFS, n);
+			}
 		}
 	}
 	else if (type == OBJ_SECTORS)
 	{
 		// delete the sidedefs bound to this sector
-		for (int n = NumSideDefs-1; n >= 0; n--)
+		for (int n = NumSideDefs-1 ; n >= 0 ; n--)
 		{
 			if (SideDefs[n]->sector == objnum)
+			{
 				BA_Delete(OBJ_SIDEDEFS, n);
+			}
 		}
 	}
+
+	SYS_ASSERT(cur_group);
+
+	cur_group->Add_Apply(op);
 }
 
 
-bool BA_Change(obj_type_t type, int objnum, byte field, int value)
+bool BA_Change(obj_type_e type, int objnum, byte field, int value)
 {
+	// TODO: optimise, check whether value actually changes
+
 	edit_op_c op;
 
 	op.action  = OP_CHANGE;
@@ -741,6 +925,7 @@ bool BA_Change(obj_type_t type, int objnum, byte field, int value)
 	SYS_ASSERT(cur_group);
 
 	cur_group->Add_Apply(op);
+	return true;
 }
 
 
@@ -749,12 +934,16 @@ bool BA_Undo()
 	if (undo_history.empty())
 		return false;
 
+	DoClearChangeStatus();
+
 	undo_group_c * grp = undo_history.front();
 	undo_history.pop_front();
 
 	grp->ReApply();
 
 	redo_future.push_front(grp);
+
+	DoProcessChangeStatus();
 	return true;
 }
 
@@ -763,12 +952,16 @@ bool BA_Redo()
 	if (redo_future.empty())
 		return false;
 	
+	DoClearChangeStatus();
+
 	undo_group_c * grp = redo_future.front();
 	redo_future.pop_front();
 
 	grp->ReApply();
 
 	undo_history.push_front(grp);
+
+	DoProcessChangeStatus();
 	return true;
 }
 
@@ -777,24 +970,25 @@ void BA_ClearAll()
 {
 	int i;
 
-	for (i = 0; i < NumThings;   i++) delete Things[i];
-	for (i = 0; i < NumVertices; i++) delete Vertices[i];
-	for (i = 0; i < NumSectors;  i++) delete Sectors[i];
-	for (i = 0; i < NumSideDefs; i++) delete SideDefs[i];
-	for (i = 0; i < NumLineDefs; i++) delete LineDefs[i];
-	for (i = 0; i < NumRadTrigs; i++) delete RadTrigs[i];
+	for (i = 0 ; i < NumThings   ; i++) delete Things[i];
+	for (i = 0 ; i < NumVertices ; i++) delete Vertices[i];
+	for (i = 0 ; i < NumSectors  ; i++) delete Sectors[i];
+	for (i = 0 ; i < NumSideDefs ; i++) delete SideDefs[i];
+	for (i = 0 ; i < NumLineDefs ; i++) delete LineDefs[i];
 
 	Things.clear();
 	Vertices.clear();
 	Sectors.clear();
 	SideDefs.clear();
 	LineDefs.clear();
-	RadTrigs.clear();
-
-	basis_strtab.clear();
 
 	ClearUndoHistory();
 	ClearRedoFuture();
+
+	// Note: we don't clear the string table, since there can be
+	//       string references in the clipboard.
+
+	Clipboard_ClearLocals();
 }
 
 
@@ -803,7 +997,7 @@ void BA_ClearAll()
 bool BA_ChangeTH(int thing, byte field, int value)
 {
 	SYS_ASSERT(is_thing(thing));
-	SYS_ASSERT(field <= Thing::F_OPTIONS);
+	SYS_ASSERT(field <= Thing::F_ARG5);
 
 	return BA_Change(OBJ_THINGS, thing, field, value);
 }
@@ -835,18 +1029,10 @@ bool BA_ChangeSD(int side, byte field, int value)
 bool BA_ChangeLD(int line, byte field, int value)
 {
 	SYS_ASSERT(is_linedef(line));
-	SYS_ASSERT(field <= LineDef::F_LEFT);
+	SYS_ASSERT(field <= LineDef::F_ARG5);
 
 	return BA_Change(OBJ_LINEDEFS, line, field, value);
 }
 
-bool BA_ChangeRAD(int rad, byte field, int value)
-{
-	SYS_ASSERT(is_radtrig(rad));
-	SYS_ASSERT(field <= RadTrig::F_CODE);
-
-	return BA_Change(OBJ_RADTRIGS, rad, field, value);
-}
-
 //--- editor settings ---
 // vi:ts=4:sw=4:noexpandtab
