@@ -4,7 +4,7 @@
 //
 //  Eureka DOOM Editor
 //
-//  Copyright (C) 2001-2009 Andrew Apted
+//  Copyright (C) 2001-2013 Andrew Apted
 //  Copyright (C) 1997-2003 André Majorel et al
 //
 //  This program is free software; you can redistribute it and/or
@@ -24,16 +24,38 @@
 //
 //------------------------------------------------------------------------
 
+#ifndef __EUREKA_E_LINEDEF_H__
+#define __EUREKA_E_LINEDEF_H__
+
 class bitvec_c;
 #include "selectn.h"
 
-bitvec_c *bv_vertices_of_linedefs (bitvec_c *linedefs);
-bitvec_c *bv_vertices_of_linedefs (SelPtr list);
-SelPtr list_vertices_of_linedefs (SelPtr list);
+void FlipLineDef(int ld);
+void FlipLineDefGroup(selection_c& flip);
 
-void centre_of_linedefs (selection_c * list, int *x, int *y);
 
-void frob_linedefs_flags (selection_c * list, int op, int operand);
+void LineDefs_SetLength(int new_len);
+
+bool LineDefAlreadyExists(int v1, int v2);
+bool LineDefWouldOverlap(int v1, int x2, int y2);
+
+void DeleteLineDefs(selection_c *lines);
+
+int SplitLineDefAtVertex(int ld, int v_idx);
+
+void MoveCoordOntoLineDef(int ld, int *x, int *y);
+
+void LineDefs_Align(int ld, int side, int sd, char part, const char *flags);
+
+
+/* commands */
+
+void LIN_Flip(void);
+void LIN_Align(void);
+void LIN_MergeTwo(void);
+void LIN_SplitHalf(void);
+
+#endif  /* __EUREKA_E_LINEDEF_H__ */
 
 //--- editor settings ---
 // vi:ts=4:sw=4:noexpandtab
