@@ -24,10 +24,24 @@
 //
 //------------------------------------------------------------------------
 
+#ifndef __EUREKA_E_VERTEX_H__
+#define __EUREKA_E_VERTEX_H__
 
-void centre_of_vertices (selection_c * list, int *x, int *y);
-void centre_of_vertices (const bitvec_c &bv, int &x, int &y);
+void MergeVertex(int v1, int v2, bool v1_will_be_deleted);
 
+void InsertPolygonVertices (int, int, int, int);
+
+int VertexHowManyLineDefs(int v_num);
+void Vertex_MergeList(selection_c *list);
+int Vertex_FindExact(int x, int y);
+
+void VERT_Merge(void);
+void VERT_Disconnect(void);
+
+void LIN_Disconnect(void);
+void SEC_Disconnect(void);
+
+#endif  /* __EUREKA_E_VERTEX_H__ */
 
 //--- editor settings ---
 // vi:ts=4:sw=4:noexpandtab
