@@ -4,7 +4,7 @@
 //
 //  Eureka DOOM Editor
 //
-//  Copyright (C) 2001-2009 Andrew Apted
+//  Copyright (C) 2001-2013 Andrew Apted
 //  Copyright (C) 1997-2003 André Majorel et al
 //
 //  This program is free software; you can redistribute it and/or
@@ -24,17 +24,35 @@
 //
 //------------------------------------------------------------------------
 
-
-#ifndef YH_X_HOVER  /* DO NOT INSERT ANYTHING BEFORE THIS LINE */
-#define YH_X_HOVER
+#ifndef __EUREKA_X_HOVER_H__
+#define __EUREKA_X_HOVER_H__
 
 
 class Objid;
 
+void GetCurObject(Objid& o, obj_type_e objtype, int x, int y, bool snap = false);
 
-void GetCurObject (Objid& o, int objtype, int x, int y);
+void GetSplitLineDef(Objid& o, int x, int y, int drag_vert = -1);
+
+float ApproxDistToLineDef(const LineDef * L, int x, int y);
+
+int ClosestLine_CastingHoriz(int x, int y, int *side);
+int ClosestLine_CastingVert (int x, int y, int *side);
+int ClosestLine_CastAtAngle (int x, int y, float radians);
+
+int OppositeLineDef(int ld, int ld_side, int *result_side);
+int OppositeSector(int ld, int ld_side);
+
+void FastOpposite_Begin();
+void FastOpposite_Finish();
+
+bool PointOutsideOfMap(int x, int y);
+
+// result: -1 for back, +1 for front, 0 for _exactly_on_ the line
+int PointOnLineSide(int x, int y, int lx1, int ly1, int lx2, int ly2);
 
 
-#endif  /* DO NOT ADD ANYTHING AFTER THIS LINE */
+#endif  /* __EUREKA_X_HOVER_H__ */
+
 //--- editor settings ---
 // vi:ts=4:sw=4:noexpandtab
